/*-
 * Copyright (C)2012 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2012 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: switch.c,v 1.7 2013/05/01 16:51:18 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#ifdef __FreeBSD__
# include <babolo/BLINflag.h>
#else
# include "linux.h"
#endif

#include "transposit.h"

#define TOBY(A) (((A) + 7) >> 3)

int
/*****************************************************************************************************
 *    ,    0.                            *
 *****************************************************************************************************/
transposit_swcmp(transposit_switch *swr0, transposit_switch *swr1) {
    int ex = 0;
    size_t i;

    if  (swr0->nou < swr1->nou) {
        ex = -1;
        goto out;
    } else if (swr0->nou > swr1->nou) {
        ex = 1;
        goto out;
    }
    for (i = 0; i < swr0->nou; ++i) {
        if  (swr0->swt[i] < swr1->swt[i]) {
            ex = -1;
            goto out;
        } else if (swr0->swt[i] > swr1->swt[i]) {
            ex = 1;
            goto out;
    }   }
out:
    return(ex);
}

int
/*****************************************************************************************************
 *     swr   sdp,                                     *
 *    codes  coden .                                            *
 *      ,      .             *
 *****************************************************************************************************/
transposit_swdepo(transposit_deposit **sdp, transposit_switch *swr, void *codes, u_char coden) {
    transposit_depocodes *dc;
    transposit_deposit *dp;
    int ex = 0;

    for (dp = *sdp;; dp = dp->next) {
        if  (!dp) {
            if  (!(dp = calloc(1, sizeof(transposit_deposit)))) {
                errno = ENOMEM;
                ex = -EX_OSERR;
                goto out;
            }
            dp->next = *sdp;
            *sdp = dp;
            dp->swc = swr;
            if  (!(dp->codes = calloc(1, TOBY(coden) + sizeof(transposit_depocodes)))) {
                errno = ENOMEM;
                ex = -EX_OSERR;
                goto out;
            }
            dp->codes->coden = coden;
            bcopy(codes, dp->codes->codes, TOBY(coden));
            ex = 1;
            goto out;
        }
        if  (!transposit_swcmp(swr, dp->swc)) {
            for (dc = dp->codes;; dc = dc->next) {
                if  (!dc) {
                    if  (!(dc = calloc(1, sizeof(transposit_depocodes)))) {
                        errno = ENOMEM;
                        ex = -EX_OSERR;
                        goto out;
                    }
                    dc->next = dp->codes;
                    dp->codes = dc;
                    dc->coden = coden;
                    bcopy(codes, dc->codes, TOBY(coden));
                    goto out;
    }   }   }   }
out:
    return(ex);
}

static u_char n[] =
/*0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 -----------------------------------------------      */
{ 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4 /* 0 */
, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5 /* 1 */
, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5 /* 2 */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* 3 */
, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5 /* 4 */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* 5 */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* 6 */
, 3, 4, 4, 5, 4, 5, 5, 6 ,4, 5, 5, 6, 5, 6, 6, 7 /* 7 */
, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5 /* 8 */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* 9 */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* A */
, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7 /* B */
, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6 /* C */
, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7 /* D */
, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7 /* E */
, 4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 /* F */
};

static u_char m[] =
/*  0     1     2     3     4     5     6     7     8 */
{ 0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF
};

static int
codeval(transposit_depocodes *dc, void *codes, u_char coden) {
    u_char *c = codes;
    int i, l, ex = 0;

    for (i = 0, l = (dc->coden < coden) ? dc->coden : coden; l > 0; l-=8, ++i) {
        ex += n[~(dc->codes[i] ^ c[i]) & m[(l > 8) ? 8 : l]];
    }
    return(ex);
}

int
/*****************************************************************************************************
 *      
 *       .
 *     codes .
 *****************************************************************************************************/
transposit_swuniq(transposit_deposit *sdp, void *codes, u_char coden) {
    transposit_depocodes **dn, **dc = NULL;
    transposit_deposit *dp;
    int t, f, ex = 0;

    for (dp = sdp; dp; dp = dp->next) {
        for (t = -1, dn = &dp->codes; *dn;) {
            transposit_depocodes *ddd;

            f = codeval(*dn, codes, coden);
            if  (t < 0) {
                t = f;
                dc = dn;
                dn = &(*dn)->next;
            } else if (f <= t) {
                /*      */
                ddd = *dn;
                *dn = ddd->next;
                free(ddd);
            } else {
                /*      */
                t = f;
                ddd = *dc;
                *dc = *dn;
                dn = &(*dn)->next;
                free(ddd);
    }   }   }
    return(ex);
}

int
/*****************************************************************************************************
 *     .
 *****************************************************************************************************/
transposit_swsort(transposit_deposit **sdp) {
    transposit_deposit **dp, *dt;
    int i = 1, ex = 0;

    while (i) {
        for (i = 0, dp = sdp; *dp && (*dp)->next;) {
            if  (0 < transposit_swcmp((*dp)->swc, (*dp)->next->swc)) {
                ++i;
                dt = *dp;
                *dp = (*dp)->next;
                dt->next = (*dp)->next;
                (*dp)->next = dt;
            } else {
                dp = &(*dp)->next;
    }   }   }
    return(ex);
}

u_int64_t
/*****************************************************************************************************
 *  swr   ,  width   .                                            *
 *  width == 0,    .                                              *
 *****************************************************************************************************/
transposit_pack(transposit_switch *swr, int width) {
    u_int64_t l, m;
    size_t i;

    if  (!width) for (i = swr->nin - 1; i; i >>= 1) ++width;
    for (m = 0, i = 0; i < width; ++i) m = (m << 1) | 1;
    for (l = 0, i = 0; i < swr->nou; ++i) l |= (u_int64_t)(swr->swt[i] & m) << (i * width);
    return(l);
}

void
/*****************************************************************************************************
 *        swr   o.                             *
 *      .                                                             *
 *         , *  ,                    *
 *     .                                                                *
 *****************************************************************************************************/
transposit_swprint(transposit_switch *swr, FILE *o) {
    size_t i, l;

    fprintf(o, "      ");
    for (l = 0; l < swr->nin; ++l) fprintf(o, "%2"BLIN_U, BLIN_I(l));
    fprintf(o, "\n");
    for (i = 0; i < swr->nou; ++i) {
        fprintf(o, "   %2"BLIN_U":", BLIN_I(i));
        for (l = 0; l < swr->nin; ++l) {
            fprintf(o, " %c", (l == swr->swt[i]) ? '*' : '.');
        }
        fprintf(o, "\n");
}   }

void
/*****************************************************************************************************
 *        o,    :       *
 *  -          ;                                 *
 *  -      transposit_swprint()                                *
 *****************************************************************************************************/
transposit_deprint(transposit_deposit *sdp, FILE *o) {
    transposit_depocodes *dc;
    size_t i, l;

    for (i = 0; sdp; ++i, sdp = sdp->next) {
        fprintf(o, "##%5"BLIN_U":", BLIN_I(i));
        for (dc = sdp->codes; dc; dc = dc->next) {
            fprintf(o, " ");
            for (l = 0; l < TOBY(dc->coden); ++l) fprintf(o, "%02X", dc->codes[l]);
        }
        fprintf(o, " #\n");
        transposit_swprint(sdp->swc, o);
}   }
